/* wptW32API.cpp - Common W32 API functions
 *	Copyright (C) 2001, 2002, 2003, 2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <shellapi.h>
#include <shlobj.h>
#include <commctrl.h>
#include <time.h>

#include "wptNLS.h"
#include "wptW32API.h"
#include "wptErrors.h"
#include "wptVersion.h"
#include "wptTypes.h"


extern "C" void _SHFree (void *p);


static void
set_menu_text_ext (HMENU menu, int by_pos, int m_uid, const char *text)
{
    MENUITEMINFO mii;
    
    memset (&mii, 0, sizeof mii);
    mii.cbSize = sizeof mii;
    mii.fMask = MIIM_TYPE;
    mii.fType = MFT_STRING;
    mii.dwTypeData = (char *) text;
    SetMenuItemInfo (menu, m_uid, by_pos? TRUE : FALSE, &mii);
}


/* Set the text of a menu item @m_uid to @text. */
void
set_menu_text (HMENU menu, UINT m_uid, const char *text)
{
    set_menu_text_ext (menu, 0, m_uid, text);
}


/* Set the text of a menu item with the position @pos to @text. */
void
set_menu_text_bypos (HMENU menu, UINT pos, const char *text)
{
    set_menu_text_ext (menu, 1, pos, text);
}


/* Set the state of a menu item @m_uid to @state. */
void
set_menu_state (HMENU menu, UINT m_uid, UINT state)
{	
    MENUITEMINFO mii;

    memset (&mii, 0, sizeof (mii));
    mii.cbSize = sizeof (mii);
    mii.fMask = MIIM_STATE;
    mii.fState = state;
    SetMenuItemInfo (menu, m_uid, FALSE, &mii);
}


/* Retrieve the state of the menu item @m_uid and return it. */
UINT
get_menu_state (HMENU menu, UINT m_uid)
{
    MENUITEMINFO mii;

    memset (&mii, 0, sizeof (mii));
    mii.cbSize = sizeof (mii);
    mii.fMask = MIIM_STATE;
    GetMenuItemInfo (menu, m_uid, FALSE, &mii);
    return mii.fState;
}


enum {
    CDLG_FILE_OPEN = 0,
    CDLG_FILE_SAVE = 1
};

/* Use the common dialog to request a file from the user. 
   id can be either FILE_OPEN or FILE_SAVE. 
   The return value is the file name or NULL if cancel was chosen. */
const char *
get_filename_dlg (HWND hwnd, int id, const char * title, 
		  const char * filter, const char * name)
{
    static char file[512] = "";
    OPENFILENAME open;
    
    if (name && strlen (name) < (sizeof (file)-1))
	strcpy (file, name);
    else
	memset (file, 0, sizeof (file));
    if (!filter)
	filter = _("All Files (*.*)\0*.*\0\0");
    /* XXX: problem with gettext because of the 'artificial' 
            double string termination!. */
    memset (&open, 0, sizeof (open));
    open.lStructSize = sizeof (OPENFILENAME);
    open.hInstance = glob_hinst;
    open.lpstrTitle = title;
    open.lpstrFilter = filter;
    open.hwndOwner = hwnd;
    open.lpstrFile = file;
    open.nMaxFile = sizeof (file) - 1;
    if (id == CDLG_FILE_OPEN)
	open.Flags = OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
    else
	open.Flags = OFN_OVERWRITEPROMPT;

    if (id == CDLG_FILE_OPEN && GetOpenFileName (&open))
	return open.lpstrFile;
    else if (id == CDLG_FILE_SAVE && GetSaveFileName (&open))
	return open.lpstrFile;
    
    return NULL;
}

const char*
get_filesave_dlg (HWND hwnd, const char *title,
		  const char *filter, const char *name)
{
    return get_filename_dlg (hwnd, CDLG_FILE_SAVE, title, filter, name);
}

const char*
get_fileopen_dlg (HWND hwnd, const char *title, const char *filter,
		  const char *name)
{
    return get_filename_dlg (hwnd, CDLG_FILE_OPEN, title, filter, name);
}


/* Use the common dialog to allow the user to select a folder.
   The return value is either the folder path or NULL if cancel was chosen. */
const char*
get_folder_dlg (HWND hwnd, const char *title, const char *name)
{
    static char folder[MAX_PATH+1] = "";
    BROWSEINFO bi;
    ITEMIDLIST *il;

    memset (&bi, 0, sizeof (bi));
    bi.hwndOwner = hwnd;
    if (title)
	bi.lpszTitle = title;
    if (name && strlen (name) < MAX_PATH-1)
	strcpy (folder, name);
    else
	memset (folder, 0, sizeof (folder));
    il = SHBrowseForFolder (&bi);
    if (il) {
	SHGetPathFromIDList (il, folder);
	_SHFree (il);
	return folder;
    }
    return NULL;
}


/* Return the clipboard contents as a string or NULL
   if the clipboard does not contain text. */
char*
get_clip_text (HWND hwnd)
{
    HANDLE clipmem;
    char *cliptxt, *p;  
    int len;
    
    if (OpenClipboard (hwnd) == FALSE)
	return NULL;
    clipmem = GetClipboardData (CF_TEXT);
    if (clipmem == NULL) {
	p = NULL;
	goto leave;
    }
    cliptxt = (char *) GlobalLock (clipmem);
    if (cliptxt == NULL) {
	p = NULL;
	goto leave;
    }
    
    len =  strlen (cliptxt);
    p = new char[len + 1];
    if (!p)
	BUG (NULL);
    memcpy (p, cliptxt, len);
    p[len] = '\0';
    GlobalUnlock (clipmem);
    
leave:
    CloseClipboard ();
    return p;
}


/* Set @text as the new clipboard content. */
int
set_clip_text (HWND hwnd, const char *text, int nbytes)
{    
    HANDLE clipmem;
    int rc = 0;
    char *p;
    
    if (OpenClipboard (hwnd) == FALSE)	
	return WPTERR_CLIP_OPEN;
    EmptyClipboard ();

    clipmem = GlobalAlloc (GHND, nbytes + 1);
    if (clipmem == NULL)
	BUG (NULL);
    p = (char *) GlobalLock (clipmem);
    if (p == NULL) {
	rc = WPTERR_GENERAL;;
	goto leave;	
    }
    memcpy (p, text, nbytes);
    p[nbytes] = '\0';
    
    GlobalUnlock (clipmem);
    SetClipboardData (CF_TEXT, clipmem);
    
leave:
    CloseClipboard ();
    return rc;
} /* set_clip_text */


/* Append or prepend some text to the clipboard contents.
   If as_footer = 1, append the text otherwise prepend. */
int
set_clip_text2 (HWND hwnd, const char *text, int nbytes, int as_footer)
{
    char *p, *new_text;

    p = get_clip_text (hwnd);
    if (!p)
	return WPTERR_CLIP_GET;
    new_text = new char [strlen (p)+strlen (text)+8];
    if (!new_text)
	BUG (0);
    if (as_footer == 0)
	sprintf (new_text, "%s\r\n%s\r\n\r\n", text, p);
    else
	sprintf (new_text, "%s\n%s\n\n", p, text);
    set_clip_text (hwnd, new_text, strlen (new_text)+1);
    free_if_alloc (p);
    free_if_alloc (new_text);
    return 0;
}


/* Make a file name out of the path @path, the file @file and 
   an extension. @ext.
   Return value: the full file name on success. */
char*
make_filename (const char *path, const char *file, const char *ext)
{
    char *p;
    size_t size = 0;

    if( path && *path )
	size += strlen( path );
    if( file && *file )
	size += strlen( file );
    if( ext && *ext )
	size += strlen( ext );
    p = new char[size + 4];
    memset( p, 0, size );
    if( path ) {
	strcat( p, path );
	if( path[strlen( path ) -1] != '\\' )
	    strcat( p, "\\" );
    }
    if( file )
	strcat( p, file );
    if( ext ) {
	strcat( p, "." );
	strcat( p, ext );
    }
    return p;
}


/* Generate a file name from a special dirctory. */
char*
make_special_filename (int folder, const char *file, const char *ext)
{
    BOOL ec;
    char path[MAX_PATH], *p;
    size_t n=0;

    /* MSDN: buf must be at least MAX_PATH=256 bytes */
    ec = SHGetSpecialFolderPath (HWND_DESKTOP, path, folder, TRUE);
    if (ec != 1) {
	log_debug ("SHGetSpecialFolderPath() failed\r\n", (int)GetLastError ());
	return NULL;
    }

    n = strlen (path)+1;
    if (file)
	n += strlen (file)+1;
    if (ext)
	n += strlen (ext)+1;
    p = new char[n+2];
    if (!p)
	BUG (0);
    memset (p, 0, n+2);
    strcpy (p, path);    
    if (file) {
	strcat (p, "\\");
	strcat (p, file);
    }
    if (ext)
	strcat (p, ext);
    return p;
}


/* return 0 if the file @fname exists, otherwise >0. */
int
file_exist_check (const char *fname)
{
    struct stat st;
    if (stat (fname, &st) == -1)
	return WPTERR_FILE_EXIST;
    return 0;
}


/* Check if the current folder exists.
   Return 0 for success. */
int
dir_exist_check (const char *dir)
{
    struct stat statbuf;    
    
    if (stat (dir, &statbuf) == -1)
	return WPTERR_GENERAL;
    if (statbuf.st_mode & _S_IFDIR)
	return 0;
    return WPTERR_GENERAL;
}


/* Return the file size of the given file @fname. */
DWORD
get_file_size (const char *fname)
{
    DWORD fsize;
    HANDLE fh;

    fh = CreateFile (fname, GENERIC_READ, FILE_SHARE_READ,
		     NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (fh == INVALID_HANDLE_VALUE)
	return 0;
    fsize = GetFileSize (fh, NULL);
    if (fsize == 0xFFFFFFFF)
	fsize = 0;
    CloseHandle (fh);
    return fsize;
}


/* Start a dialog with the exception that before it is checked that the
   dialog is not already openened. */
int
dialog_box_param (HINSTANCE hinst, LPCTSTR name, HWND parent, DLGPROC fnc, 
		  LPARAM param, LPCTSTR title, int title_id)
{
    if (FindWindowEx (GetDesktopWindow (), NULL, NULL, title))
	return -1;
    return DialogBoxParam (hinst, name, parent, fnc, param);
}


/* Wrapper for message box which forces the message box into the
   foreground and it is displayed always on top. */
int
msg_box (HWND hwnd, const char *text, const char *title, int mode)
{
    mode |= MB_SETFOREGROUND;
    mode |= MB_TASKMODAL;
    mode |= MB_TOPMOST;
    return MessageBox(hwnd, text, title, mode);
}


/* Safe strdup version (C++ version). */
char*
m_strdup (const char *str)
{
    char *p = new char[strlen (str) + 1];
    if (!p)
	BUG (NULL);
    strcpy (p, str);
    return p;
}


/* Center the hwndChild relative to parent.
   The style param allows to specificy additional styles (like topmost). */
void
center_window2 (HWND hwndChild, HWND parent, HWND style) 
{     
    HWND hwndParent;
    RECT rChild, rParent;    
    HDC hdc;
    int wChild, hChild, wParent, hParent;     
    int wScreen, hScreen, xNew, yNew;
    int flags = SWP_NOSIZE | SWP_NOZORDER;

    hwndParent = parent;
    if (hwndParent == NULL)
	hwndParent = GetDesktopWindow ();
    GetWindowRect (hwndChild, &rChild);     
    wChild = rChild.right - rChild.left;     
    hChild = rChild.bottom - rChild.top;

    GetWindowRect (hwndParent, &rParent);     
    wParent = rParent.right - rParent.left;     
    hParent = rParent.bottom - rParent.top;      
    
    hdc = GetDC (hwndChild);     
    wScreen = GetDeviceCaps (hdc, HORZRES);     
    hScreen = GetDeviceCaps (hdc, VERTRES);     
    ReleaseDC (hwndChild, hdc);      
    xNew = rParent.left + ((wParent - wChild) /2);     
    if (xNew < 0)
	xNew = 0;
    else if ((xNew+wChild) > wScreen) 
	xNew = wScreen - wChild;
    yNew = rParent.top  + ((hParent - hChild) /2);
    if (yNew < 0)
	yNew = 0;
    else if ((yNew+hChild) > hScreen)
	yNew = hScreen - hChild;
    if (style == HWND_TOPMOST || style == HWND_NOTOPMOST)
	flags = SWP_NOMOVE | SWP_NOSIZE;
    SetWindowPos (hwndChild, style? style : NULL, xNew, yNew, 0, 0, flags);
}


/* Center the given hwndChild window with no special style. */
void
center_window (HWND hwndChild, HWND hwndParent) 
{
    center_window2 (hwndChild, hwndParent, NULL);
}


/* Retrieve the product verion of the given file @fname.
   Format: MAJOR.MINOR.PATCH1.PATCH2
   Return value: 0 on success. */
int
get_file_version (const char *fname, WORD *major, WORD *minor, 
		  WORD *patch1, WORD *patch2)
{
    VS_FIXEDFILEINFO *inf = {0};
    char file[MAX_PATH+1] = {0};
    LPVOID buf, data;
    DWORD arg;
    DWORD size;
    UINT qlen;

    strncpy (file, fname, MAX_PATH);
    size = GetFileVersionInfoSize (file, &arg);
    if (!size)
	return -1;
    buf = (LPVOID)new CHAR[size];
    if (!buf)
	BUG (NULL);
    GetFileVersionInfo (file, 0, size, buf);

    qlen=0;
    VerQueryValue (buf, "\\", &data, &qlen);
    if (!qlen) {
	delete [] (char*)buf;
	return -1;
    }
    inf = (VS_FIXEDFILEINFO*)data;

    if (major)
	*major = HIWORD (inf->dwProductVersionMS);
    if (minor)
	*minor = LOWORD (inf->dwProductVersionMS);
    if (patch1)
	*patch1 = HIWORD (inf->dwProductVersionLS);	
    if (patch2)
	*patch2 = LOWORD (inf->dwProductVersionLS);

    delete [] (char*)buf;
    return 0;
}


/* Return date in a format which complies with the
   system locale settings. */
const char*
get_locale_date (long tm_t, char *buf, DWORD buflen)
{ 
    SYSTEMTIME st; 
    struct tm *ptm;

    ptm = localtime (&tm_t);  
    st.wYear = (WORD)ptm->tm_year;
    st.wMonth = (WORD)ptm->tm_mon;
    st.wDay = (WORD)ptm->tm_mday;
    st.wYear += 1900;
    st.wMonth += 1;
    if (!GetDateFormat (LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, 
		        NULL, buf, buflen))
	return NULL;
    return buf;
}


struct reminder_hd_s {
    int msecs;
    HWND dlg;
    HANDLE hd;
};


static DWORD CALLBACK
foreground_reminder_thread (void *c)
{
    struct reminder_hd_s *ctx = (struct reminder_hd_s *)c;
    Sleep (ctx->msecs);
    SetForegroundWindow (ctx->dlg);
    CloseHandle (ctx->hd);
    delete ctx;
    ExitThread (0);
    return 0;
}

/* Try to force the window @dlg to the foreground.
   On NT5 or later this will not work if the user
   is working in another window (console for example). */
void
force_foreground_window (HWND dlg, int msecs)
{
    struct reminder_hd_s *hd;
    DWORD tid;

    hd = new reminder_hd_s;
    hd->dlg = dlg;
    hd->msecs = msecs;
    hd->hd = CreateThread (NULL, 0, foreground_reminder_thread, 
			   hd, NULL, &tid);
}
